#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import os
import tempfile
from importlib import metadata as importlib_metadata
from pathlib import Path

import tomli


class BaseConfig(object):
    """Base configuration class

    This class holds and/or gathers all the APIs base settings...

    """

    APP_NAME = os.environ.get("APP_NAME", __name__)
    APP_DESCRIPTION = os.environ.get("APP_DESCRIPTION", """Alliander Weather Provider API""")
    APP_MAINTAINER = os.environ.get("APP_MAINTAINER", "DNB/ST Innovatieteam")
    APP_MAINTAINER_EMAIL = os.environ.get("APP_MAINTAINER_EMAIL", "weather.provider@alliander.com")
    SHOW_MAINTAINER = os.environ.get("SHOW_MAINTAINER", False)

    default_version = None
    try:
        if Path("./pyproject.toml").exists():
            pyproject_file = Path("./pyproject.toml")
        elif Path("../pyproject.toml").exists():
            pyproject_file = Path("../pyproject.toml")
        with open(pyproject_file, mode="rb") as file:
            default_version = tomli.load(file)["tool"]["poetry"]["version"]
    except Exception:
        default_version = importlib_metadata.version(__package__)
    finally:
        if not default_version:
            default_version = "Version Unidentified"
    APP_VERSION = os.environ.get("APP_VERSION", default_version)

    APP_V1_VERSION = os.environ.get("APP_V1_VERSION", f"{APP_VERSION} - v1 API (1.0.3)")
    APP_V2_VERSION = os.environ.get("APP_V1_VERSION", f"{APP_VERSION} - v2 API (2.0.3)")
    APP_VALID_DATE = os.environ.get("APP_VALID_DATE", "2024-12-31")
    NETWORK_INTERFACE = os.environ.get("NETWORK_INTERFACE", "127.0.0.1")
    NETWORK_PORT = os.environ.get("NETWORK_PORT", 8080)
    LOG_LEVEL = os.environ.get("LOG_LEVEL", "DEBUG")

    # The following two parameters determine the active configuration
    DEPLOYED = os.environ.get("DEPLOYED", False)
    DEBUG = os.environ.get("DEBUG", True)

    REPO_FOLDER = os.environ.get("REPO_FOLDER", f"{tempfile.gettempdir()}/Weather_Repository")


class LocalConfig(BaseConfig):
    pass


class LocalDebugConfig(LocalConfig):
    DEBUG = True
    LOG_LEVEL = os.environ.get("LOG_LEVEL", "DEBUG")


class DeployedConfig(BaseConfig):
    NETWORK_INTERFACE = os.environ.get("NETWORK_INTERFACE", "0.0.0.0")
    DEPLOYED = True


class DeployedDebugConfig(DeployedConfig):
    DEBUG = True
    LOG_LEVEL = os.environ.get("LOG_LEVEL", "DEBUG")


_active_config = None


def get_active_config():  # pragma: no cover
    """Retrieve the active configuration.

    The active configuration is based on the DEPLOYED and DEBUG
    environment variables. The active config is set once and returned
    for each subsequent call.

    Returns:
        Active configuration
    """
    global _active_config

    is_deployed = os.environ.get("DEPLOYED", False)
    is_debug = os.environ.get("DEBUG", False)

    if _active_config is None:
        if is_deployed:
            if is_debug:
                _active_config = DeployedDebugConfig
            else:
                _active_config = DeployedConfig
        else:
            if is_debug:
                _active_config = LocalDebugConfig
            else:
                _active_config = LocalConfig

    return _active_config


def get_setting(key: str):
    """Return value set for key.

    Args:
        key (str): key of configuration setting

    Returns:
        value if key exists, None otherwise
    """
    active_config = get_active_config()
    return getattr(active_config, key, None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Utilities for handling KNMI datasets.
"""
from typing import List

import numpy as np
import pandas as pd
from geopy.distance import great_circle

from weather_provider_api.routers.weather.utils.geo_position import GeoPosition


def find_closest_stn_list(stn_stations: pd.DataFrame, coords: List[GeoPosition]):
    """
        A function that finds the closest stations to the locations in the given list of GeoPositions
    Args:
        stn_stations:   A Pandas Dataframe containing all the station data for the KNMI stations
        coords:         A list of GeoPositions with locations to find the nearest stations to.
    Returns:
        A list of coordinates for found stations
        A list of found stations in the same order
        A list of index numbers for the found in the same order
    """
    _stn_stations = stn_stations.copy(deep=True)

    # get list of relevant STNs, choose closest STN
    coords_stn = [_find_closest_stn_single(_stn_stations, coord) for coord in coords]
    stns = list(set(coords_stn))
    coords_stn_ind = [stns.index(x) for x in coords_stn]
    return coords_stn, stns, coords_stn_ind


def _find_closest_stn_single(stn_stations: pd.DataFrame, coord: GeoPosition):
    """
        A function that finds the closest station to a single GeoPosition
    Args:
        stn_stations:   A Pandas Dataframe containing all the station data for the KNMI stations
        coord:          A GeoPosition to find the nearest station for
    Returns:
        A station number indicating its index in the supplied dataframe.
    """
    stn_stations["distance"] = stn_stations.apply(
        lambda x: great_circle((x["lat"], x["lon"]), coord.get_WGS84()).km, axis=1
    )

    # Find the stn with the lowest distance to the location
    min_ind = np.argmin(stn_stations["distance"].values)
    # Return the found stn
    return stn_stations.loc[min_ind, "STN"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import structlog
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse

from weather_provider_api.app_config import get_setting

logger = structlog.get_logger(__name__)


async def handle_http_exception(request: Request, exc: StarletteHTTPException) -> JSONResponse:
    headers = getattr(exc, "headers", None)

    body = {"detail": exc.detail}
    if get_setting("SHOW_MAINTAINER"):
        body["maintainer"] = get_setting("APP_MAINTAINER")
        body["maintainer_email"] = get_setting("APP_MAINTAINER_EMAIL")

    return JSONResponse(body, status_code=exc.status_code, headers=headers)


def initialize_error_handling(app):
    app.add_exception_handler(StarletteHTTPException, handle_http_exception)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request

from weather_provider_api.app_config import get_setting


def initialize_metadata_header_middleware(app):
    async def add_metadata_headers(request: Request, call_next):
        response = await call_next(request)
        response.headers["X-App-Version"] = get_setting("APP_VERSION")
        response.headers["X-App-Valid-Till"] = get_setting("APP_VALID_DATE")
        return response

    async def add_metadata_headers_with_maintainer(request: Request, call_next):
        response = await call_next(request)
        response.headers["X-App-Version"] = get_setting("APP_VERSION")
        response.headers["X-App-Valid-Till"] = get_setting("APP_VALID_DATE")
        response.headers["X-Maintainer"] = get_setting("APP_MAINTAINER")
        response.headers["X-Maintainer-Email"] = get_setting("APP_MAINTAINER_EMAIL")
        return response

    show_maintainer = get_setting("SHOW_MAINTAINER")
    if show_maintainer:
        insertion_func = add_metadata_headers_with_maintainer
    else:
        insertion_func = add_metadata_headers

    app.add_middleware(BaseHTTPMiddleware, dispatch=insertion_func)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from fastapi import APIRouter

router = APIRouter()


# TODO: implement API status retrieval
def get_api_status():
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from fastapi import FastAPI

from weather_provider_api.app_config import get_setting
from weather_provider_api.routers.weather.api_view_v1 import app as weather_router

app = FastAPI(title="Weather API", root_path="/api/v1", version=get_setting("APP_V1_VERSION"))

app.include_router(weather_router, prefix="/weather")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import datetime

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request

from weather_provider_api.app_config import get_setting
from weather_provider_api.core.errors.exceptions import APIExpiredException
from weather_provider_api.core.initializers.error_handling import handle_http_exception


def initialize_validation_middleware(app):

    valid_date = datetime.datetime.strptime(get_setting("APP_VALID_DATE"), "%Y-%m-%d")

    async def check_api_validity(request: Request, call_next):
        if datetime.datetime.now() > valid_date:
            response = await handle_http_exception(request, APIExpiredException())
        else:
            response = await call_next(request)

        return response

    app.add_middleware(BaseHTTPMiddleware, dispatch=check_api_validity)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import logging
import logging.config

import structlog

from weather_provider_api.app_config import get_setting


def add_application_metadata(logger, method_name, event_dict):
    if "app_version" not in event_dict:
        event_dict["app_version"] = get_setting("APP_VERSION")
    if "app_valid_date" not in event_dict:
        event_dict["app_valid_date"] = get_setting("APP_VALID_DATE")

    return event_dict


def initialize_logging():
    timestamper = structlog.processors.TimeStamper(fmt="iso")

    shared_processors = [
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        add_application_metadata,
        timestamper,
        structlog.processors.UnicodeDecoder(),
    ]

    pre_chain = shared_processors

    # Setup JSON-logging to stdout
    logging.captureWarnings(True)
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "json": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processor": structlog.processors.JSONRenderer(),
                    "foreign_pre_chain": pre_chain,
                }
            },
            "handlers": {
                "default": {
                    "level": get_setting("LOG_LEVEL"),
                    "class": "logging.StreamHandler",
                    "formatter": "json",
                }
            },
            "loggers": {
                "": {
                    "handlers": ["default"],
                    "level": get_setting("LOG_LEVEL"),
                    "propagate": True,
                }
            },
        }
    )

    structlog.configure(
        processors=[structlog.stdlib.filter_by_level]
        + shared_processors
        + [structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import structlog
from starlette_prometheus import PrometheusMiddleware, metrics

from weather_provider_api.app_config import get_setting

logger = structlog.get_logger(__name__)


def initialize_prometheus_middleware(app, endpoint="/metrics"):
    """Start HTTP endpoint for Prometheus monitoring.

    Args:
        app: FastAPI weather_provider_api instance
        endpoint (str): URL at which the metrics should be available
    """
    if get_setting("DEPLOYED"):
        logger.info(f'Enabling Prometheus endpoint on: "{endpoint}"')
        app.add_middleware(PrometheusMiddleware)
        app.add_route(endpoint, metrics)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from pydantic import BaseModel as PydanticBaseModel


class BaseModel(PydanticBaseModel):
    class Config:
        orm_mode = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0


"""Main API executable
This module contains everything required to boot the Weather Provider API.
Its '__main__' if-clause handles running it directly from a directory or IDE setting.

"""
from datetime import datetime

import structlog
import uvicorn
from fastapi import FastAPI
from slowapi import _rate_limit_exceeded_handler
from slowapi.errors import RateLimitExceeded
from starlette.responses import RedirectResponse

from weather_provider_api.app_config import get_setting
from weather_provider_api.core.initializers.error_handling import initialize_error_handling
from weather_provider_api.core.initializers.headers import initialize_metadata_header_middleware
from weather_provider_api.core.initializers.logging import initialize_logging
from weather_provider_api.core.initializers.monitoring import initialize_prometheus_middleware
from weather_provider_api.core.initializers.mounting import mount_api_version
from weather_provider_api.core.initializers.rate_limiter import API_RATE_LIMITER
from weather_provider_api.core.initializers.validation import initialize_validation_middleware
from weather_provider_api.versions.v1 import app as v1
from weather_provider_api.versions.v2 import app as v2

app = FastAPI(version=get_setting("APP_VERSION"), title=get_setting("APP_NAME"))

# Add rate limiter
app.state.limiter = API_RATE_LIMITER
app.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)


# Enable logging
initialize_logging()
logger = structlog.get_logger(__name__)
logger.info("--------------------------------------", datetime=datetime.utcnow())
logger.info("Booting Weather Provider API Systems..", datetime=datetime.utcnow())
logger.info("--------------------------------------", datetime=datetime.utcnow())

# Create and configure new application instance
initialize_error_handling(app)
initialize_metadata_header_middleware(app)
initialize_validation_middleware(app)
initialize_prometheus_middleware(app)

# Activate enabled API versions
mount_api_version(app, v1)
mount_api_version(app, v2)


# Redirect users to the docs
@app.get("/")
def redirect_to_docs():
    redirect_url = "/api/v2/docs"  # replace with docs URL or use weather_provider_api.url_path_for()
    return RedirectResponse(url=redirect_url)


logger.info(f"--------------------------------------", datetime=datetime.utcnow())
logger.info(f"Finished booting; starting uvicorn...", datetime=datetime.utcnow())
logger.info(f"--------------------------------------", datetime=datetime.utcnow())


def main():
    uvicorn.run(app, host="127.0.0.1", port=8080)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from fastapi import FastAPI

from weather_provider_api.core.initializers.error_handling import initialize_error_handling


def mount_api_version(base_app: FastAPI, versioned_app):
    """Mounts a versioned API application.

    This method helps to expose an API version to the main application.
    Additionally, it ensures Alliander core API endpoints are set. The latter
    are obligatory and may not be removed or bypassed.

    Args:
        base_app: main FastAPI application
        versioned_app: versioned FastAPI application
    """
    from weather_provider_api.core.views import router

    initialize_error_handling(versioned_app)
    versioned_app.include_router(router, prefix="/core")
    base_app.mount(versioned_app.root_path, versioned_app)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

""" KNMI current weather data fetcher.
"""
import copy
import locale
import re
from datetime import datetime
from typing import List, Optional

import numpy as np
import pandas as pd
import requests
import structlog
import xarray as xr

from weather_provider_api.routers.weather.base_models.model import WeatherModelBase
from weather_provider_api.routers.weather.sources.knmi.stations import (
    stations_actual,
    stations_actual_reversed,
)
from weather_provider_api.routers.weather.sources.knmi.utils import (
    find_closest_stn_list,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition
from weather_provider_api.routers.weather.utils.pandas_helpers import coords_to_pd_index

logger = structlog.get_logger(__name__)


class ActueleWaarnemingenModel(WeatherModelBase):
    """
    A Weather Model that incorporates the:
    "KNMI Actuele Waarnemingen" dataset into the Weather Provider API
    """

    def __init__(self):
        super().__init__()
        self.id = "waarnemingen"
        self.name = "KNMI Actuele Waarnemingen"
        self.version = None
        self.url = "https://www.knmi.nl/nederland-nu/weer/waarnemingen"
        self.predictive = False
        self.time_step_size_minutes = 10
        self.num_time_steps = 1
        self.description = "Current weather observations. Updated every 10 minutes."
        self.async_model = False

        self.to_si = {
            "weather_description": {
                "name": "weather_description",
                "convert": self.no_conversion,
            },
            "temperature": {"convert": self.celsius_to_kelvin},
            "humidity": {"convert": self.percentage_to_frac},
            "wind_direction": {"convert": self.dutch_wind_direction_to_degrees},
            "wind_speed": {"convert": self.no_conversion},  # m/s
            "visibility": {"convert": self.no_conversion},  # m
            "air_pressure": {"convert": lambda x: x * 100},  # hPa to Pa
        }
        self.to_human = copy.deepcopy(self.to_si)
        self.to_human["temperature"]["convert"] = self.no_conversion  # C

        self.human_to_model_specific = self._create_reverse_lookup(self.to_si)

    def get_weather(
        self,
        coords: List[GeoPosition],
        begin: Optional[np.datetime64],
        end: Optional[np.datetime64],
        weather_factors: List[str] = None,
    ) -> xr.Dataset:
        """
        The function that gathers and processes the requested Actuele Waarnemingen weather data from the KNMI site
        and returns it as a Xarray Dataset.
        (This model interprets directly from an HTML page, but the information is also available from the data
        platform. Due to it being rather impractically handled, we stick to the site for now.)

        Args:
            coords:             A list of GeoPositions containing the locations the data is requested for.
            begin:              A datetime containing the start of the period to request data for.
            end:                A datetime containing the end of the period to request data for.
            weather_factors:    A list of weather factors to request data for (in string format)

        Returns:
            An Xarray Dataset containing the weather data for the requested period, locations and factors.

        NOTES:
            As this model only return the current weather data the 'begin' and 'end' values are not actually used.
        """
        updated_weather_factors = self._request_weather_factors(weather_factors)

        # Download the current weather data
        raw_ds = self._download_weather()

        # Get a list of the relevant STNs and choose the closest STN for each coordinate
        coords_stn, stns, coords_stn_ind = find_closest_stn_list(
            stations_actual, coords
        )

        # Select the data for the found closest STNs
        ds = raw_ds.sel(STN=coords_stn)

        data_dict = {
            var_name: (["time", "coord"], var.values)
            for var_name, var in ds.data_vars.items()
            if var_name in updated_weather_factors and var_name not in ["lat", "lon"]
        }

        timeline = ds.coords["time"].values

        ds = xr.Dataset(
            data_vars=data_dict,
            coords={"time": timeline, "coord": coords_to_pd_index(coords)},
        )
        ds = ds.unstack("coord")
        return ds

    def _download_weather(self) -> Optional[xr.Dataset]:
        """
        A function that downloads the Actuele Waarnemingen site and parses it into a Weather Provider API compatible
        format.

        Returns:
            An Xarray Dataset containing the parsed weather data.
        """
        raw_ds = None

        try:
            knmi_site_response = requests.get(self.url)

            if knmi_site_response.ok:
                knmi_site_df = pd.read_html(knmi_site_response.text)[0]

                column_translations = {
                    "Station": "station",
                    "Weer": "weather_description",
                    "Temp (°C)": "temperature",
                    "Chill (°C)": "wind_chill",
                    "RV (%)": "relative_humidity",
                    "Wind (bft)": "wind_direction",
                    "Wind (m/s)": "wind_speed",
                    "Wind (km/uur)": "wind_speed_kmu",
                    "Zicht (m)": "visibility",
                    "Druk (hPa)": "air_pressure",
                }

                # Rename to the conventional naming system used for the Weather Provider API
                for dictionary_item in knmi_site_df.columns.copy(deep=True):
                    if dictionary_item in column_translations.keys():
                        knmi_site_df = knmi_site_df.rename(
                            columns={
                                dictionary_item: column_translations[dictionary_item]
                            }
                        )
                    else:
                        knmi_site_df = knmi_site_df.drop(
                            dictionary_item, axis="columns"
                        )

                current_observation_moment = self.retrieve_observation_moment(
                    knmi_site_response.text
                )
                knmi_site_df["time"] = current_observation_moment
                if "wind_direction" in knmi_site_df:
                    knmi_site_df["wind_direction"] = knmi_site_df[
                        "wind_direction"
                    ].str.strip("\n")

                # Add a field for the station with its (lat, lon)-coordinates and remove the original station code
                knmi_site_df["STN"] = knmi_site_df["station"].apply(
                    lambda x: stations_actual_reversed[x.upper()]
                )

                stations_actual_indexed = stations_actual.set_index("STN")

                knmi_site_df["lat"] = knmi_site_df["STN"].apply(
                    lambda x: stations_actual_indexed.loc[x, "lat"]
                )
                knmi_site_df["lon"] = knmi_site_df["STN"].apply(
                    lambda x: stations_actual_indexed.loc[x, "lon"]
                )
                knmi_site_df = knmi_site_df.drop("station", axis="columns")

                # Rebuild the index
                knmi_site_df = knmi_site_df.set_index(["time", "STN"])

                # The following fields get removed because they aren't currently used in this API
                if "weather_description" in knmi_site_df:
                    knmi_site_df = knmi_site_df.drop(
                        "weather_description", axis="columns"
                    )
                if "wind_chill" in knmi_site_df:
                    knmi_site_df = knmi_site_df.drop("wind_chill", axis="columns")

                # Generate a Xarray DataSet
                raw_ds = knmi_site_df.to_xarray()

        except (requests.exceptions.BaseHTTPError, IndexError) as e:
            logger.exception(str(e))

        return raw_ds

    @staticmethod
    def retrieve_observation_moment(html_body: str):
        """
        A function that extracts the observation date from the supplied html body of the Actuele Waarnemingen site.

        Args:
            html_body:  The text from the body downloaded from the Actuele Waarnemingen Site.

        Returns:
            Either the matching datetime if it can be extracted, or the current datetime if it cannot.
        """
        try:
            re_match = re.search(
                r"Waarnemingen\s(\d+\s\w+\s\d{4}\s\d{2}:\d{2})\suur",
                html_body,
            )
            if re_match:
                dt_str = re_match.group(1)
                current_locale = locale.getlocale(locale.LC_TIME)
                locale.setlocale(locale.LC_TIME, "dutch")
                dt = datetime.strptime(dt_str, "%d %B %Y %H:%M")
                locale.setlocale(locale.LC_TIME, current_locale)
                return dt
        except locale.Error as e:
            logger.warning('No locale could be determined for KNMI Waarnemingen datetime transformation. '
                           'Using local datetime instead.', datetime=datetime.utcnow())
            return datetime.now()
        except Exception as e:
            logger.exception(
                'An unknown exception occurred while retrieving the KNMI Waarnemingen datetime for use in the response. '
                f'Using local datetime instead. Error: {e}', datetime=datetime.utcnow()
            )
            return datetime.now()

        current_locale = locale.getlocale(locale.LC_TIME)
        locale.setlocale(locale.LC_TIME, "dutch")
        dt = datetime.now()
        locale.setlocale(locale.LC_TIME, current_locale)
        return dt

    def is_async(self):  # pragma: no cover
        return self.async_model

    def _request_weather_factors(self, factors: Optional[List[str]]) -> List[str]:
        # Implementation of the Base Weather Model function that returns a list of known weather factors for the model.
        if factors is None:
            return list(self.to_si.keys())

        new_factors = []

        for f in factors:
            f_low = f.lower()
            if f_low in self.to_si:
                new_factors.append(f_low)

        return list(set(new_factors))  # Cleanup any duplicate values and return
